--
-- Database creation
--

CREATE EXTENSION IF NOT EXISTS dblink;

DO $$
DECLARE
    db_name TEXT;
BEGIN

    db_name := 'pscheduler';

    IF NOT EXISTS (SELECT * FROM pg_roles WHERE rolname = db_name)
    THEN
        PERFORM dblink_exec('dbname=' || current_database(),
                            'CREATE ROLE pscheduler WITH LOGIN');
        RAISE NOTICE 'Created role %', db_name;
    END IF;

    IF NOT EXISTS (SELECT * FROM pg_database WHERE datname = db_name)
    THEN
        PERFORM dblink_exec('dbname=' || current_database(),
                            'CREATE DATABASE pscheduler WITH OWNER=pscheduler');
        PERFORM dblink_exec('dbname=' || current_database(),
                            'GRANT ALL ON DATABASE pscheduler TO pscheduler');
        RAISE NOTICE 'Created database %', db_name;
    END IF;

END;
$$ LANGUAGE plpgsql;


\c pscheduler

CREATE OR REPLACE LANGUAGE plpythonu;

-- Needed for generating UUIDs
CREATE EXTENSION IF NOT EXISTS pgcrypto;
--
-- Stored procedure that runs external programs
--

-- NOTE: This uses an untrusted language and therefore must be
-- installed as the superuser

DROP TYPE IF EXISTS external_program_result CASCADE;
CREATE TYPE external_program_result AS (
       status INTEGER,
       stdout TEXT,
       stderr TEXT
);


--
-- Run the pscheduler frontend's "internal" command
--
-- TODO: Figure out what do do when DB is off-node.  Probably need an
-- external process like the task runner.
--
-- TODO: This doesn't fail if the program never returns.  Need to do
-- something about that, maybe along the lines of
-- http://stackoverflow.com/a/4825933.  (Uses threading... Ick.)
--
--
--
-- HACK: BWCTLBC -- Can probably leave this.
--
-- The env_add argument is an array of alternating pairs of variable
-- names and values that are to be added to the environment before the
-- program runs.  [ 'FOO', 'bar', 'BAZ', 'quux' ] would be turned into
-- the Python equivalent of {"FOO": "bar", "BAZ": "quux"}.
--
CREATE OR REPLACE FUNCTION pscheduler_internal(
    argv TEXT[] DEFAULT '{}',  -- Arguments to pass to 'pscheduler internal'
    input TEXT DEFAULT NULL,   -- What goes to the standard input
    timeout INTEGER DEFAULT 5, -- How long to wait in seconds
    env_add TEXT[] DEFAULT ARRAY[]::TEXT[]  -- Environment additions.  See above.
)
RETURNS external_program_result
AS $$

import pscheduler


if env_add is None or len(env_add) == 0:
    env_add_hash = {}
else:
    if len(env_add) % 2 != 0:
        plpy.error('env_add must have an even number of elements.')
    env_add_hash = {}
    while env_add:
        env_add_hash[env_add.pop()] = env_add.pop()


try:

    status, stdout, stderr = pscheduler.run_program(
        [ 'pscheduler', 'internal' ] + argv,
	stdin = input,
	timeout = timeout,
	env_add = env_add_hash
	)

except Exception as ex:
    plpy.error('Unable to run pscheduler front-end program: ' + str(ex))

return { 'status': status, 'stdout': stdout, 'stderr': stderr }

$$ LANGUAGE plpythonu;
--
-- Stored procedures for HTTP(S) Operations
--

-- NOTE: This uses an untrusted language and therefore must be
-- installed as the superuser

-- TODO: It would be nice to use PgCURL
-- (https://github.com/Ormod/pgcurl) or something like it instead of
-- dragging this through Python.  This'll do for now.

DROP TYPE IF EXISTS http_result CASCADE;
CREATE TYPE http_result AS (
       status INTEGER,
       returned TEXT
);


CREATE OR REPLACE FUNCTION http_get(
    url TEXT DEFAULT NULL,
    timeout FLOAT default NULL,
    bind TEXT default NULL
)
RETURNS http_result
AS $$

import pscheduler

try:
    status, returned = pscheduler.url_get(url, json=False, throw=False,
        timeout=timeout, bind=bind)
except Exception as ex:
    plpy.error("Failed to GET %s: %s" % (url, str(ex)))

return { "status": status, "returned": returned }

$$ LANGUAGE plpythonu;



CREATE OR REPLACE FUNCTION http_put(
    url TEXT DEFAULT NULL,
    payload TEXT DEFAULT NULL,
    timeout FLOAT default NULL,
    bind TEXT default NULL
)
RETURNS http_result
AS $$

import pscheduler

try:
    status, returned = pscheduler.url_put(url, data=payload,
                                          json=False, throw=False,
                                          timeout=timeout, bind=bind)
except Exception as ex:
    plpy.error("Failed to PUT %s: %s" % (url, str(ex)))

return { "status": status, "returned": returned }

$$ LANGUAGE plpythonu;



CREATE OR REPLACE FUNCTION http_post(
    url TEXT DEFAULT NULL,
    payload TEXT DEFAULT NULL,
    timeout FLOAT default NULL,
    bind TEXT default NULL
)
RETURNS http_result
AS $$

import pscheduler

try:
    status, returned = pscheduler.url_post(url, data=payload,
                                           json=False, throw=False,
                                           timeout=timeout, bind=bind)
except Exception as ex:
    plpy.error("Failed to POST %s: %s" % (url, str(ex)))

return { "status": status, "returned": returned }

$$ LANGUAGE plpythonu;



CREATE OR REPLACE FUNCTION http_delete(
    url TEXT DEFAULT NULL,
    timeout FLOAT default NULL,
    bind TEXT default NULL
)
RETURNS http_result
AS $$

import pscheduler

try:
    status, returned = pscheduler.url_delete(url, throw=False,
    	    timeout=timeout, bind=bind)
except Exception as ex:
    plpy.error("Failed to DELETE %s: %s" % (url, str(ex)))

return { "status": status, "returned": returned }

$$ LANGUAGE plpythonu;
--
-- Stored procedure that validates JSON using the pScheduler library
--

-- NOTE: This uses an untrusted language and therefore must be
-- installed as the superuser

--
-- Validate JSON
--
--
CREATE OR REPLACE FUNCTION json_validate(
    candidate JSONB,                               -- JSON to be validated
    json_ref TEXT DEFAULT '#/pScheduler/AnyJSON'   -- What object to validate
)
RETURNS TEXT  -- NULL if okay, error text otherwise.
AS $$

import pscheduler

if not isinstance(candidate, basestring):
    plpy.error("No idea what to do with type %s" % (type(candidate)))

json = pscheduler.json_load(candidate)

valid, message = pscheduler.json_validate(
    {"": json},
    {    
        "type": "object",
        "properties": {
            "": { "$ref": json_ref }
        },
        "required": [ "" ]
    })

return None if valid else message

$$ LANGUAGE plpythonu;
